from verifier import Verifier
from iotile.core.exceptions import ValidationError


class OptionsVerifier(Verifier):
    """Verify that an object is one of a set of options

    Args:
        *args (Verifier): A set of possible values to compare
        desc (string): A description of the verifier
    """

    def __init__(self, *args, **kwargs):
        desc = kwargs.get('desc', None)
        super(OptionsVerifier, self).__init__(desc)

        self._options = args

    def verify(self, obj):
        """Verify that the object conforms to this verifier's schema

        Args:
            obj (object): A python object to verify

        Raises:
            ValidationError: If there is a problem verifying the dictionary, a
                ValidationError is thrown with at least the reason key set indicating
                the reason for the lack of validation.
        """

        if len(self._options) == 0:
            raise ValidationError("No options", reason='no options given in options verifier, matching not possible', object=obj)

        exceptions = {}

        for i, option in enumerate(self._options):
            try:
                option.verify(obj)
                return
            except ValidationError, exc:
                exceptions['option_%d' % (i+1)] = exc.params['reason']

        raise ValidationError("Object did not match any of a set of options", reason="object did not match any given option (first failure = '%s')" % exceptions['option_1'], **exceptions)



# This file is adapted from python code released by WellDone International
# under the terms of the LGPLv3.  WellDone International's contact information is
# info@welldone.org
# http://welldone.org
#
# Modifications to this file from the original created at WellDone International 
# are copyright Arch Systems Inc.

#formatting.py

def indent_block(string_block, level):
	indent = ' '*level
	repstr = '\n' + indent

	retval = string_block.replace('\n', repstr)
	return indent + retval

def indent_list(list, level):
	"""
	Join a list of strings, one per line with 'level' spaces before each one
	"""

	indent = ' '*level
	joinstr = '\n' + indent

	retval = joinstr.join(list)
	return indent + retval


from ws4py.client.threadedclient import WebSocketClient
from iotile.core.hw.exceptions import *
from iotile.core.exceptions import *
from iotile.core.hw.commands import RPCCommand
from iotile.core.hw.reports.parser import IOTileReportParser
import threading
from Queue import Queue, Empty
from cmdstream import CMDStream
import msgpack
import datetime
import time


class WSIOTileClient(WebSocketClient):
    def __init__(self, port, report_callback):
        super(WSIOTileClient, self).__init__(port, protocols=['iotile-ws'])

        self.connection_established = threading.Event()
        self.messages = Queue()
        self.report_callback = report_callback

    def start(self):
        try:
            self.connect()
        except Exception, exc:
           raise HardwareError("Unable to connect to websockets host", reason=str(exc))

        self.connection_established.wait()

    def opened(self):
        self.connection_established.set()

    def closed(self, code, reason):
        self.connection_established.clear()

    def unpack(self, msg):
        return msgpack.unpackb(msg, object_hook=self.decode_datetime)

    @classmethod
    def decode_datetime(cls, obj):
        if b'__datetime__' in obj:
            obj = datetime.datetime.strptime(obj[b'as_str'].decode(), "%Y%m%dT%H:%M:%S.%f")
        return obj

    @classmethod
    def encode_datetime(cls, obj):
        if isinstance(obj, datetime.datetime):
            obj = {'__datetime__': True, 'as_str': obj.strftime("%Y%m%dT%H:%M:%S.%f").encode()}
        return obj

    def received_message(self, message):
        unpacked_message = self.unpack(message.data)

        if 'type' in unpacked_message and unpacked_message['type'] == 'report':
            report = IOTileReportParser.DeserializeReport(unpacked_message['value'])
            self.report_callback(report)
        else:
            self.messages.put(unpacked_message)


class WebSocketStream(CMDStream):
    """An IOTile commander transport stream using web sockets to a server.

    The server is responsible for implementing direct communication with the IOTile devices.
    This stream just provides a generic way to connect to a websockets based server.

    If connection_string is not None, a connection will be made immediately to a specific IOTile
    device.  Otherwise, no connection will be made and devices can be discovered by using the
    scan routine.

    Args:
        port (str): A url for the websocket server in form of server:port/path
        connection_string (str): a protocol specific connection string indicating a specific device
    """

    def __init__(self, port, connection_string, record=None):
        port = "ws://{0}".format(port)
        self._report_queue = Queue()
        self.client = WSIOTileClient(port, report_callback=self._report_callback)
        self.client.start()
        self._connection_id = None

        CMDStream.__init__(self, port, connection_string, record=record)

    def _close(self):
        if self.client.connection_established.is_set():
            self.client.close()

    def _scan(self):
        result = self.send('scan')
        return result['devices']

    def _connect(self, uuid_value):
        args = {}
        args['uuid'] = uuid_value

        result = self.send('connect', args)
        self._connection_id = result['connection_id']

        connstring = result['connection_string']

        try:
            result = self.send('open_interface', {'interface': 'rpc'})
        except:
            self._disconnect()
            raise

        return connstring

    def _connect_direct(self, connection_string):
        args = {}
        args['connection_string'] = connection_string

        result = self.send('connect_direct', args)
        self._connection_id = result['connection_id']

        try:
            result = self.send('open_interface', {'interface': 'rpc'})
        except:
            self._disconnect()
            raise

    def _enable_streaming(self):
        self.send('open_interface', {'interface': 'streaming'})
        return self._report_queue

    def _send_rpc(self, address, feature, cmd, payload, **kwargs):
        args = {}
        args['rpc_address'] = address
        args['rpc_feature'] = feature
        args['rpc_command'] = cmd
        args['rpc_payload'] = str(payload)

        timeout = 3.0
        if 'timeout' in kwargs:
            timeout = float(kwargs['timeout'])

        args['rpc_timeout'] = timeout

        result = self.send('send_rpc', args)
        status = result['status']
        payload = result['payload']

        return status, payload

    def _disconnect(self):
        self.send('disconnect')

    def _report_callback(self, report):
        self._report_queue.put(report)

    def send(self, command, args={}, progress=None):
        cmd = {}
        cmd['command'] = command
        cmd.update(args)

        self.client.send(msgpack.packb(cmd), binary=True)

        done = False

        while not done:
            try:
                result = self.client.messages.get(timeout=10.0)
            except Empty:
                raise TimeoutError('Timeout waiting for response to %s command from websocket server' % command)

            if 'type' in result and result['type'] == 'progress':
                total = result['total']
                current = result['current']

                if progress:
                    progress(current, total)
            else:
                done = True

        if result['success'] != True:
            raise HardwareError("Error processing command", reason=result['reason'])

        return result

    def _send_highspeed(self, data, progress_callback):
        self.send('send_script', {'data': str(data)}, progress=progress_callback)



"""Classes for parsing and dealing with semantic versions and ranges of versions
"""

from functools import total_ordering
from iotile.core.exceptions import DataError, ArgumentError

@total_ordering
class SemanticVersion(object):
    """A simple class representing a version in X.Y.Z[-prerelease] format

    Only a known number of prerelease types are allowed and must be in the format:
    <name><integer>

    where name is one of: build, alpha, beta, rc
    and integer is a whole number greater than 0

    The sort order for prereleases is defined as build < alpha < beta < rc < release
    """

    prerelease_order = {'build': 0, 'alpha': 1, 'beta': 2, 'rc': 3, 'release': 4}

    def __init__(self, major, minor, patch, release_type='release', prerelease_number=None):
        self.major = major
        self.minor = minor
        self.patch = patch
        self.release_type = release_type
        self.prerelease_number = prerelease_number

    def inc_first_nonzero(self):
        """Create a new SemanticVersion with the first nonzero value incremented

        If this version is 0.0.1, then this will return 0.0.2.  If this version
        if 0.1.0 then this version will return 0.2.0.  

        All prerelease information is stripped so 1.0.0-alpha2 becomes 2.0.0.
        """

        release = [x for x in self.release_tuple]

        if release[0] == 0 and release[1] == 0:
            release[2] += 1
        elif release[0] == 0:
            release[1] += 1
            release[2] = 0
        else:
            release[0] += 1
            release[1] = 0
            release[2] = 0

        return SemanticVersion(*release)

    @property
    def coexistence_class(self):
        """A tuple representing the compatibility class of this semantic version

        Coexistence classes are defined as a tuple containing the release version
        with everything except the first nonzero entry being zero.  Basically,
        coexistence classes divide version into those that should be compatible
        based on semantic versioning rules 
        """

        out = self.release_tuple

        if out[0] == 0 and out[1] == 0:
            return out
        elif out[0] == 0:
            return (out[0], out[1], 0)
        else:
            return (out[0], 0, 0)

    def inc_release(self):
        """Create a new SemanticVersion with the patch level incremented

        All prerelease information is stripped.  So 0.0.1-alpha2 becomes 0.0.2
        """

        release = [x for x in self.release_tuple]
        release[2] += 1

        return SemanticVersion(*release)

    @property
    def is_release(self):
        return self.release_type == 'release'

    @property
    def is_prerelease(self):
        return not self.is_release

    @classmethod
    def ParsePrerelease(cls, prerelease):
        """Parse a prerelease string into a type, number tuple
        
        Args:
            prerelease (string): a prerelease string in the format specified for SemanticVersion

        Returns:
            tuple: (release_type, number)
        """

        last_alpha = 0
        while last_alpha < len(prerelease) and prerelease[last_alpha].isalpha():
            last_alpha += 1

        release_type = prerelease[:last_alpha]
        release_number = prerelease[last_alpha:]

        if release_type not in SemanticVersion.prerelease_order or release_type == 'release':
            raise DataError("Invalid Prerelease specifier in semantic version", prerelease_type=release_type)

        try:
            release_number = int(release_number)
        except ValueError:
            raise DataError("Invalid Prerelease number in semantic version", prerelease_number=release_number)

        return (release_type, release_number)

    @classmethod
    def FromString(cls, version):
        parts = version.split('.')
        if len(parts) != 3:
            raise DataError("Invalid version format in SemanticVersion, must be X.Y.Z[-prerelease]", version=version)

        major = int(parts[0])
        minor = int(parts[1])

        if '-' in parts[2]:
            patchstr, prerelease = parts[2].split('-')
            patch = int(patchstr)

            release_type, prerelease_num = cls.ParsePrerelease(prerelease)
        else:
            patch = int(parts[2])
            release_type = 'release'
            prerelease_num = None

        return SemanticVersion(major, minor, patch, release_type, prerelease_num)

    @property
    def release_tuple(self):
        """Return the X.Y.Z prefix for this version
        """

        return (self.major, self.minor, self.patch)

    def _ordering_tuple(self):
        return (self.major, self.minor, self.patch, self.prerelease_order[self.release_type], self.prerelease_number)

    def __str__(self):
        version = "{0}.{1}.{2}".format(self.major, self.minor, self.patch)

        if self.release_type is not 'release':
            version += '-{0}{1}'.format(self.release_type, self.prerelease_number)

        return version

    def __eq__(self, other):
        return self._ordering_tuple() == other._ordering_tuple()

    def __lt__(self, other):
        return self._ordering_tuple() < other._ordering_tuple()

    def __hash__(self):
        return hash(self._ordering_tuple())


class SemanticVersionRange(object):
    """Class specifying a range of SemanticVersion objects

    Ranges can be used to filter a list of SemanticVersion objects into
    only those that satisfy the range conditions.  Currently, only a minimal
    set of range operations is supported.  Range parsing and interpretation
    is taken from npm semver:

    https://docs.npmjs.com/misc/semver

    Currently, the only operations that are implemented in the parser is:

    ^X.Y.Z

    The preferred way to create a SemanticVersionRange is through the classmethod
    FromString("<version range spec>").

    Args:
        disjuncts (list): A list of lists of 4 tuples that specify the ranges to be matched
            Each tuple should (lower, upper, inclusive_lower, inclusive_upper) where
            lower and upper and SemanticVersions (possibly None) and inclusive_* are
            bools that determine whether the range condition is <= upper or < upper 
            (>= lower, > lower).  Each sublist is joined conjunctively within itself
            and the lists themselves are joined disjunctively.
    """

    def __init__(self, disjuncts):
        self._disjuncts = disjuncts

    def _check_ver_range(self, version, ver_range):
        """Check if version is included in ver_range
        """
        
        lower, upper, lower_inc, upper_inc = ver_range

        #If the range extends over everything, we automatically match
        if lower is None and upper is None:
            return True

        if lower is not None:
            if lower_inc and version < lower:
                return False
            elif not lower_inc and version <= lower:
                return False

        if upper is not None:
            if upper_inc and version > upper:
                return False
            elif not upper_inc and version >= upper:
                return False

        #Prereleases have special matching requirements
        if version.is_prerelease:
            #Prereleases cannot match ranges that are not defined as prereleases
            if (lower is None or not lower.is_prerelease) and (upper is None or not upper.is_prerelease):
                return False

            #Prereleases without the same major.minor.patch as a range end point cannot match
            if (lower is not None and version.release_tuple != lower.release_tuple) and \
               (upper is not None and version.release_tuple != upper.release_tuple):
                return False

        return True

    def _check_insersection(self, version, ranges):
        """Check that a version is inside all of a list of ranges
        """

        for ver_range in ranges:
            if not self._check_ver_range(version, ver_range):
                return False

        return True

    def check(self, version):
        """Check that a version is inside this SemanticVersionRange
        
        Args:
            version (SemanticVersion): The version to check

        Returns:
            bool: True if the version is included in the range, False if not
        """

        for disjunct in self._disjuncts:
            if self._check_insersection(version, disjunct):
                return True

        return False

    def filter(self, versions, key=lambda x: x):
        """Filter all of the versions in an interable that match this version range

        Args:
            versions (iterable): An iterable of SemanticVersion objects

        Returns:
            list: A list of the SemanticVersion objects that matched this range
        """

        return [x for x in versions if self.check(key(x))]

    @classmethod
    def FromString(cls, range_string):
        """Parse a version range string into a SemanticVersionRange

        Currently, the only possible range strings are:

        ^X.Y.Z - matches all versions with the same leading nonzero digit
            greater than or equal the given range.
        * - matches everything
        Args:
            range_string (string): A string specifying the version range

        Returns:
            SemanticVersionRange: The resulting version range object

        Raises:
            ArgumentError: if the range string does not define a valid range.
        """

        disjuncts = None

        range_string = range_string.strip()

        if len(range_string) == 0:
            raise ArgumentError("You must pass a finite string to SemanticVersionRange.FromString", range_string=range_string)

        #Check for *
        if len(range_string) == 1 and range_string[0] == '*':
            conj = (None, None, True, True)
            disjuncts = [[conj]]

        #Check for ^X.Y.Z
        elif range_string[0] == '^':
            ver = range_string[1:]

            try:
                ver = SemanticVersion.FromString(ver)
            except DataError, err:
                raise ArgumentError("Could not parse ^X.Y.Z version", parse_error=str(err), range_string=range_string)

            lower = ver
            upper = ver.inc_first_nonzero()

            conj = (lower, upper, True, False)
            disjuncts = [[conj]]

        if disjuncts is None:
            raise ArgumentError("Invalid range specification that could not be parsed", range_string=range_string)

        return SemanticVersionRange(disjuncts)



# This file is adapted from python code released by WellDone International
# under the terms of the LGPLv3.  WellDone International's contact information is
# info@welldone.org
# http://welldone.org
#
# Modifications to this file from the original created at WellDone International 
# are copyright Arch Systems Inc.

from iotile.core.hw.hwmanager import HardwareManager
from iotile.core.hw.exceptions import *
from iotile.core.exceptions import *
import unittest
import pytest
import os.path
import os

class TestHardwareManager(unittest.TestCase):
    """
    Test to make sure that the HardwareManager is working
    """

    def setUp(self):
        self.hw = HardwareManager('none')

    def tearDown(self):
        pass

    def test_unknown_module(self):
        with pytest.raises(UnknownModuleTypeError):
            self.hw._create_proxy('UnknownTileBusModule', 8)

    def test_proxy_loading(self):
        try:
            self.hw._create_proxy('EmptyProxy', 8)
        except UnknownModuleTypeError:
            pass
        except:
            raise

        num = self.hw.add_proxies(os.path.join(os.path.dirname(__file__), 'empty_proxy.py'))
        assert num == 1

        proxy = self.hw._create_proxy('EmptyProxy', 8)
        assert proxy.random_method() == True

    def test_wrongpath_proxy(self):
        with pytest.raises(ValidationError):
            self.hw.add_proxies(os.path.join(os.path.dirname(__file__), 'empty_proxy'))



# This file is adapted from python code released by WellDone International
# under the terms of the LGPLv3.  WellDone International's contact information is
# info@welldone.org
# http://welldone.org
#
# Modifications to this file from the original created at WellDone International
# are copyright Arch Systems Inc.


import os.path
import platform

def settings_directory():
    """
    Find a per user settings directory that is appropriate for each
    type of system that we are installed on.
    """

    system = platform.system()

    basedir = None

    if system == 'Windows':
        if 'APPDATA' in os.environ:
            basedir = os.environ['APPDATA']
    elif system == 'Darwin':
        basedir = os.path.expanduser('~/Library/Preferences')

    #If we're not on Windows or Mac OS X, assume we're on some
    #kind of posix system where the appropriate place would be
    #~/.config
    if basedir is None:
        basedir = os.path.expanduser('~')
        basedir = os.path.join(basedir, '.config')

    settings_dir = os.path.abspath(os.path.join(basedir, 'IOTile-Core'))
    return settings_dir



import hashlib
import hmac
import struct
import os
from auth_provider import AuthProvider, KnownSignatureMethods
from iotile.core.exceptions import NotFoundError

class EnvAuthProvider(AuthProvider):
    """Basic authentication provider that can sign and verify using user keys

    Keys must be defined in environment variables with the naming scheme:
    USER_KEY_ABCDEFGH where ABCDEFGH is the device uuid in hex prepended with 0s
    and expressed in capital letters.  So the device with UUID 0xab would look
    for the environment variable USER_KEY_000000AB.  

    The key must be a 64 character hex string that is decoded to create a 32 byte key.
    """

    @classmethod
    def _get_key(cls, device_id):
        """Attempt to get a user key from an environment variable
        """

        var_name = "USER_KEY_{0:08X}".format(device_id)

        if var_name not in os.environ:
            raise NotFoundError("No user key could be found for devices", device_id=device_id, expected_variable_name=var_name)

        key_var = os.environ[var_name]
        if len(key_var) != 64:
            raise NotFoundError("User key in variable is not the correct length, should be 64 hex characters", device_id=device_id, key_value=key_var)

        try:
            key = key_var.decode('hex')
        except ValueError:
            raise NotFoundError("User key in variable could not be decoded from hex", device_id=device_id, key_value=key_var)

        if len(key) != 32:
            raise NotFoundError("User key in variable is not the correct length, should be 64 hex characters", device_id=device_id, key_value=key_var)

        return key

    def sign(self, device_id, sig_method, data):
        """Sign a buffer of data on behalf of a device

        This routine only supports user key based signing

        Args:
            device_id (int): The id of the device that we should encrypt for
            sig_method (int): The method of encryption that we should perform
            data (bytearray): The data that we should sign

        Returns:
            dict: The signature and any associated metadata about the signature.
                The signatured itself must always be a bytearray stored under the 
                'signature' key, however additional keys may be present depending 
                on the signature method used.

        Raises:
            NotFoundError: If the auth provider is not able to sign the data.
        """

        self._check_signature_method(sig_method)
        method_name = KnownSignatureMethods[sig_method]

        if method_name == 'hmac_sha256_user_key':
            key = self._get_key(device_id)

            #We sign the SHA256 hash of the message
            message_hash = hashlib.sha256(data).digest()
            hmac_calc = hmac.new(key, message_hash, hashlib.sha256)
            result = bytearray(hmac_calc.digest())
        else:
            raise NotFoundError('unsupported signature method in EnvAuthProvider', method=method_name)

        return {'signature': result, 'method': method_name}

    def verify(self, device_id, sig_method, data, signature):
        """Verify the signature attached to a buffer of data

        Args:
            device_id (int): The id of the device that we should encrypt for
            sig_method (int): The method of signing that was used
            data (bytearray): The data whose signature we should verify
            signature (bytearray): The signature attached to data

        Returns:
            dict: The result of the verification process must always be a bool under the
                'verified' key, however additional keys may be present depending on the 
                signature method used.

        Raises:
            NotFoundError: If the auth provider is not able to verify the data because the method
                is not supported.
        """

        self._check_signature_method(sig_method)
        method_name = KnownSignatureMethods[sig_method]

        if method_name == 'hmac_sha256_user_key':
            key = self._get_key(device_id)
            message_hash = hashlib.sha256(data).digest()
            hmac_calc = hmac.new(key, message_hash, hashlib.sha256)
            result = bytearray(hmac_calc.digest())
        else:
            raise NotFoundError('unsupported signature method in EnvAuthProvider', method=method_name)

        if len(signature) == 0:
            verified = False
        elif len(signature) > len(result):
            verified = False
        elif len(signature) < len(result):
            trunc_result = result[:len(signature)]
            verified = hmac.compare_digest(signature, trunc_result)
        else:
            verified = hmac.compare_digest(signature, result)

        return {'verified': verified, 'bit_length': 8*len(signature)}
 


"""An adapter class that takes a DeviceAdapter and produces a CMDStream compatible interface
"""
from copy import deepcopy
import Queue
from cmdstream import CMDStream
import datetime
import time
from iotile.core.exceptions import HardwareError
from iotile.core.utilities.typedargs import iprint

class AdapterCMDStream(CMDStream):
    """An adapter class that takes a DeviceAdapter and produces a CMDStream compatible interface
    
    DeviceAdapters have a more generic interface that is not restricted to exclusive use in an online
    fashion by a single user at a time.  This class implements the CMDStream interface on top of a
    DeviceAdapter.
    
    Args:
        adapter (DeviceAdapter): the DeviceAdatper that we should use to create this CMDStream
        port (string): the name of the port that we should connect through
        connection_string (string): A DeviceAdapter specific string specifying a device that we
            should immediately connect to
        record (string): The path to a file that we should use to record everything sent down
            this CMDStream
    """

    def __init__(self, adapter, port, connection_string, record=None):
        self.adapter = adapter
        self._scanned_devices = {}
        self._reports = None
        self._traces = None
        self.connection_interrupted = False

        self.adapter.add_callback('on_scan', self._on_scan)
        self.adapter.add_callback('on_report', self._on_report)
        self.adapter.add_callback('on_trace', self._on_trace)
        self.adapter.add_callback('on_disconnect', self._on_disconnect)

        self.start_time = time.time()
        self.min_scan = self.adapter.get_config('minimum_scan_time', 0.0)

        super(AdapterCMDStream, self).__init__(port, connection_string, record)

    def _on_scan(self, adapter_id, info, expiration_time):
        """Callback called when a new device is discovered on this CMDStream

        Args:
            adapter_id (int): An ID for the adapter that scanned the device
            info (dict): Information about the scanned device
            expiration_time (float): How long this device should stay around
        """

        device_id = info['uuid']
        infocopy = deepcopy(info)

        infocopy['expiration_time'] = datetime.datetime.now() + datetime.timedelta(seconds=expiration_time)
        self._scanned_devices[device_id] = infocopy

    def _on_disconnect(self, adapter_id, connection_id):
        """Callback when a device is disconnected unexpectedly

        Args:
            adapter_id (int): An ID for the adapter that was connected to the device
            connection_id (int): An ID for the connection that has become disconnected
        """

        self.connection_interrupted = True

    def _scan(self):
        elapsed = time.time() - self.start_time
        if elapsed < self.min_scan:
            time.sleep(self.min_scan - elapsed)

        to_remove = set()

        now = datetime.datetime.now()

        for name, value in self._scanned_devices.iteritems():
            if value['expiration_time'] < now:
                to_remove.add(name)

        for name in to_remove:
            del self._scanned_devices[name]

        return self._scanned_devices.values()

    def _connect(self, uuid_value):
        elapsed = time.time() - self.start_time
        if elapsed < self.min_scan:
            time.sleep(self.min_scan - elapsed)

        if uuid_value not in self._scanned_devices:
            raise HardwareError("Could not find device to connect to by UUID", uuid=uuid_value)

        connstring = self._scanned_devices[uuid_value]['connection_string']
        self._connect_direct(connstring)
        
        return connstring

    def _connect_direct(self, connection_string):
        res = self.adapter.connect_sync(0, connection_string)
        if not res['success']:
            self.adapter.periodic_callback()
            raise HardwareError("Could not connect to device", reason=res['failure_reason'], connection_string=connection_string)

        try:
            res = self.adapter.open_interface_sync(0, 'rpc')
        except Exception as exc:
            self.adapter.disconnect_sync(0)
            self.adapter.periodic_callback()
            raise HardwareError("Could not open RPC interface on device due to an exception", exception=str(exc))

        if not res['success']:
            self.adapter.disconnect_sync(0)
            self.adapter.periodic_callback()
            raise HardwareError("Could not open RPC interface on device", reason=res['failure_reason'], connection_string=connection_string)

    def _disconnect(self):
        self.adapter.disconnect_sync(0)
        self.adapter.periodic_callback()

    def _send_rpc(self, address, feature, cmd, payload, **kwargs):
        timeout = 3.0
        if 'timeout' in kwargs:
            timeout = float(kwargs['timeout'])

        result = self.adapter.send_rpc_sync(0, address, (feature << 8) | cmd, payload, timeout)
        success = result['success']
        status = result['status']
        payload = result['payload']

        #If the rpc caused us to lose connection, we will have a connection_interrupted flag set
        try:
            if self.connection_interrupted:
                self.connected = False
                self._connect_direct(self.connection_string)
                self.connection_interrupted = False
                self.connected = True

                #Reenable streaming interface if that was open before as well
                if self._reports is not None:
                    res = self.adapter.open_interface_sync(0, 'streaming')
                    if not res['success']:
                        raise HardwareError("Could not open streaming interface to device", reason=res['failure_reason'])
        except HardwareError, exc:
            raise HardwareError("Device disconnected before we received an RPC response and we could not reconnect", reconnect_error=exc)

        if not success:
            raise HardwareError("Could not send RPC", reason=result['failure_reason'])

        return status, payload

    def _send_highspeed(self, data, progress_callback):
        self.adapter.send_script_sync(0, str(data), progress_callback)

    def _enable_streaming(self):
        self._reports = Queue.Queue()
        res = self.adapter.open_interface_sync(0, 'streaming')
        if not res['success']:
            raise HardwareError("Could not open streaming interface to device", reason=res['failure_reason'])

        return self._reports

    def _enable_tracing(self):
        self._traces = Queue.Queue()
        res = self.adapter.open_interface_sync(0, 'tracing')
        if not res['success']:
            raise HardwareError("Could not open tracing interface to device", reason=res['failure_reason'])

        return self._traces

    def _on_report(self, conn_id, report):
        if self._reports is None:
            return

        self._reports.put(report)

    def _on_trace(self, conn_id, tracing_data):
        if self._traces is None:
            return

        self._traces.put(tracing_data)

    def _close(self):
        self.adapter.stop_sync()



# This file is adapted from python code released by WellDone International
# under the terms of the LGPLv3.  WellDone International's contact information is
# info@welldone.org
# http://welldone.org
#
# Modifications to this file from the original created at WellDone International 
# are copyright Arch Systems Inc.

def convert(arg):
	if arg is None:
		return None

	return str(arg)

def validate_list(arg, choices):
	"""
	Make sure the argument is in the list of choices passed to the function
	"""
	
	choice_set = set(choices)

	if arg not in choices:
		raise ValueError('Value not in list: %s' % str(choices))

def validate_not_empty(arg):
	"""
	Make sure the string is not empty
	"""

	if len(arg) == 0:
		raise ValueError("String cannot be empty")

def default_formatter(arg, **kwargs):
	return arg

def format_repr(arg, **kwargs):
	return repr(arg)


# This file is adapted from python code released by WellDone International
# under the terms of the LGPLv3.  WellDone International's contact information is
# info@welldone.org
# http://welldone.org
#
# Modifications to this file from the original created at WellDone International 
# are copyright Arch Systems Inc.

#bool.py
#Simple boolean type

def convert(arg, **kwargs):
	if arg is None:
		return arg

	if isinstance(arg, basestring):
		comp = arg.lower()
		if comp == 'true':
			return True
		elif comp == 'false':
			return False
		else:
			raise ValueError("Unknown boolean value (should be true or false): %s" % arg)

	return bool(arg)

def default_formatter(arg, **kwargs):
	return str(arg)



import hashlib
import hmac
import struct
from auth_provider import AuthProvider, KnownSignatureMethods
from iotile.core.exceptions import NotFoundError

class BasicAuthProvider(AuthProvider):
    """Basic default authentication provider that can only calculate sha256 hashes
    """

    def sign(self, device_id, sig_method, data):
        """Sign a buffer of data on behalf of a device

        This routine only supports hash only signatures

        Args:
            device_id (int): The id of the device that we should encrypt for
            sig_method (int): The method of encryption that we should perform
            data (bytearray): The data that we should sign

        Returns:
            dict: The signature and any associated metadata about the signature.
                The signatured itself must always be a bytearray stored under the 
                'signature' key, however additional keys may be present depending 
                on the signature method used.

        Raises:
            NotFoundError: If the auth provider is not able to sign the data.
        """

        self._check_signature_method(sig_method)
        method_name = KnownSignatureMethods[sig_method]

        if method_name == 'hash_only_sha256':
            result = bytearray(hashlib.sha256(data).digest())
        else:
            raise NotFoundError('unsupported signature method in BasicAuthProvider', method=method_name)

        return {'signature': result, 'method': method_name}

    def verify(self, device_id, sig_method, data, signature):
        """Verify the signature attached to a buffer of data

        Args:
            device_id (int): The id of the device that we should encrypt for
            sig_method (int): The method of signing that was used
            data (bytearray): The data whose signature we should verify
            signature (bytearray): The signature attached to data

        Returns:
            dict: The result of the verification process must always be a bool under the
                'verified' key, however additional keys may be present depending on the 
                signature method used.

        Raises:
            NotFoundError: If the auth provider is not able to verify the data because the method
                is not supported.
        """

        self._check_signature_method(sig_method)
        method_name = KnownSignatureMethods[sig_method]

        if method_name == 'hash_only_sha256':
            result = bytearray(hashlib.sha256(data).digest())
        else:
            raise NotFoundError('unsupported signature method in BasicAuthProvider', method=method_name)

        if len(signature) == 0:
            verified = False
        elif len(signature) > len(result):
            verified = False
        elif len(signature) < len(result):
            trunc_result = result[:len(signature)]
            verified = hmac.compare_digest(signature, trunc_result)
        else:
            verified = hmac.compare_digest(signature, result)

        return {'verified': verified, 'bit_length': 8*len(signature)}



"""Mock IOTile device class for testing other components interactions with IOTile devices
"""

import struct
import inspect
import binascii
from iotile.core.exceptions import IOTileException, InternalError
from iotile.core.utilities.stoppable_thread import StoppableWorkerThread
from iotile.core.hw.reports.individual_format import IndividualReadingReport
from iotile.core.hw.reports.report import IOTileReading


class RPCNotFoundError(IOTileException):
    """Exception thrown when an RPC is not found
    """
    pass


class RPCInvalidArgumentsError(IOTileException):
    """Exception thrown when an RPC with a fixed parameter format has invalid arguments
    """
    pass


class RPCInvalidReturnValueError(IOTileException):
    """Exception thrown when the return value of an RPC does not conform to its known format
    """
    pass


class RPCInvalidIDError(IOTileException):
    """Exception thrown when an RPC is created with an invalid RPC id
    """
    pass


class TileNotFoundError(IOTileException):
    """Exception thrown when an RPC is sent to a tile that does not exist
    """
    pass


def rpc(address, rpc_id, arg_format, resp_format=None):
    """Decorator to denote that a function implements an RPC with the given ID and address

    The underlying function should be a member function that will take
    individual parameters after the RPC payload has been decoded according
    to arg_format.

    Arguments to the function are decoded from the 20 byte RPC argument payload according
    to arg_format, which should be a format string that can be passed to struct.unpack.

    Similarly, the function being decorated should return an iterable of results that
    will be encoded into a 20 byte response buffer by struct.pack using resp_format as
    the format string.

    The RPC will respond as if it were implemented by a tile at address ``address`` and
    the 16-bit RPC id ``rpc_id``.

    Args:
        address (int): The address of the mock tile this RPC is for
        rpc_id (int): The number of the RPC
        arg_format (string): a struct format code (without the <) for the
            parameter format for this RPC
        resp_format (string): an optional format code (without the <) for
            the response format for this RPC
    """

    if rpc_id < 0 or rpc_id > 0xFFFF:
        raise RPCInvalidIDError("Invalid RPC ID: {}".format(rpc_id))

    def _rpc_wrapper(func):
        def _rpc_executor(self, payload):
            try:
                args = struct.unpack("<{}".format(arg_format), payload)
            except struct.error as exc:
                raise RPCInvalidArgumentsError(str(exc), arg_format=arg_format, payload=binascii.hexlify(payload))

            resp = func(self, *args)
            if resp_format is not None:
                try:
                    return struct.pack("<{}".format(resp_format), *resp)
                except struct.error as exc:
                    raise RPCInvalidReturnValueError(str(exc), resp_format(resp_format), resp=repr(resp))

            return resp

        _rpc_executor.rpc_id = rpc_id
        _rpc_executor.rpc_addr = address
        _rpc_executor.is_rpc = True
        return _rpc_executor

    return _rpc_wrapper


class VirtualIOTileDevice(object):
    """A Virtual IOTile device that can be interacted with as if it were a real one

    Args:
        iotile_id (int): A 32-bit integer that specifies the globally unique ID
            for this IOTile device.
        name (string): The 6 byte name that should be returned when anyone asks
            for the controller's name of this IOTile device using an RPC
    """

    def __init__(self, iotile_id, name):
        self._rpc_handlers = {}
        self.tiles = set()
        self.name = name
        self.iotile_id = iotile_id
        self.pending_data = False
        self.reports = []
        self.traces = []
        self.script = bytearray()

        self.connected = False
        self.stream_iface_open = False

        # For this device to push streams or tracing data through a VirtualInterface, it
        # needs access to that interface's push channel
        self._push_channel = None
        self._workers = []
        self._started = False

        # Iterate through all of our member functions and see the ones that are
        # RPCS and add them to the RPC handler table
        for name, value in inspect.getmembers(self, predicate=inspect.ismethod):
            if hasattr(value, 'is_rpc'):
                self.register_rpc(value.rpc_addr, value.rpc_id, value)

    def create_worker(self, func, interval, *args, **kwargs):
        """Spawn a worker thread running func

        The worker will be automatically be started when start() is called
        and terminated when stop() is called on this VirtualIOTileDevice.
        This must be called only from the main thread, not from a worker thread.

        create_worker must not be called after stop() has been called.  If it
        is called before start() is called, the thread is started when start()
        is called, otherwise it is called immediately.

        Args:
            func (callable): Either a function that will be called in a loop
                with a sleep of interval seconds with *args and **kwargs or
                a generator function that will be called once and expected to
                yield periodically so that the worker can check if it should
                be killed.
            interval (float): The time interval between invocations of func.
                This should not be 0 so that the thread doesn't peg the CPU
                and should be short enough so that the worker checks if it
                should be killed in a timely fashion.
            *args: Arguments that are passed to func as positional args
            **kwargs: Arguments that are passed to func as keyword args
        """

        thread = StoppableWorkerThread(func, interval, args, kwargs)
        self._workers.append(thread)

        if self._started:
            thread.start()

    def start(self, channel):
        """Start running this virtual device including any necessary worker threads

        Args:
            channel (IOTilePushChannel): the channel with a stream and trace
                routine for streaming and tracing data through a VirtualInterface
        """

        if self._started:
            raise InternalError("The method start() was called twice on VirtualIOTileDevice.")

        self._push_channel = channel
        self._started = True

        for worker in self._workers:
            worker.start()

    def stream(self, report):
        """Stream a report asynchronously

        If no one is listening for the report, the report may be dropped,
        otherwise it will be queued for sending

        Args:
            report (IOTileReport): The report that should be streamed
        """

        if self._push_channel is None:
            return

        self._push_channel.stream(report)

    def stream_realtime(self, stream, value):
        """Stream a realtime value as an IndividualReadingReport

        If the streaming interface of the VirtualInterface this
        VirtualDevice is attached to is not opened, the realtime
        reading may be dropped.

        Args:
            stream (int): The stream id to send
            value (int): The stream value to send
        """

        if not self.stream_iface_open:
            return

        reading = IOTileReading(0, stream, value)

        report = IndividualReadingReport.FromReadings(self.iotile_id, [reading])
        self.stream(report)

    def trace(self, data):
        """Trace data asynchronously

        If no one is listening for traced data, it will be dropped
        otherwise it will be queued for sending.

        Args:
            data (bytearray, string): Unstructured data to trace to any
                connected client.
        """

        if self._push_channel is None:
            return

        self._push_channel.trace(data)

    def stop(self):
        """Synchronously stop this virtual device and any potential workers
        """

        self._started = False

        for worker in self._workers:
            worker.stop()

    def register_rpc(self, address, rpc_id, func):
        """Register an RPC handler with the given info

        Args:
            address (int): The address of the mock tile this RPC is for
            rpc_id (int): The number of the RPC
            func (callable): The function that should be called to handle the
                RPC.  func is called as func(payload) and must return a single
                string object of up to 20 bytes with its response
        """

        if rpc_id < 0 or rpc_id > 0xFFFF:
            raise RPCInvalidIDError("Invalid RPC ID: {}".format(rpc_id))

        self.tiles.add(address)

        code = (address << 8) | rpc_id
        self._rpc_handlers[code] = func

    def call_rpc(self, address, rpc_id, payload=""):
        """Call an RPC by its address and ID

        Args:
            address (int): The address of the mock tile this RPC is for
            rpc_id (int): The number of the RPC
            payload (string): A byte string of payload parameters up to 20 bytes

        Returns:
            string: The response payload from the RPC
        """
        if rpc_id < 0 or rpc_id > 0xFFFF:
            raise RPCInvalidIDError("Invalid RPC ID: {}".format(rpc_id))

        if address not in self.tiles:
            raise TileNotFoundError("Unknown tile address, no registered handler", address=address)

        code = (address << 8) | rpc_id
        if code not in self._rpc_handlers:
            raise RPCNotFoundError("address: {}, rpc_id: {}".format(address, rpc_id))

        return self._rpc_handlers[code](payload)

    def open_rpc_interface(self):
        """Called when someone opens an RPC interface to the device
        """

        pass

    def close_rpc_interface(self):
        """Called when someone closes an RPC interface to the device
        """

        pass

    def open_script_interface(self):
        """Called when someone opens a script interface on this device
        """

        pass

    def open_streaming_interface(self):
        """Called when someone opens a streaming interface to the device

        Returns:
            list: A list of IOTileReport objects that should be sent out
                the streaming interface.
        """

        self.stream_iface_open = True

        return self.reports

    def open_tracing_interface(self):
        """Called when someone opens a tracing interface to the device

        Returns:
            list: A list of bytearray objects that should be sent out
                the tracing interface.
        """

        return self.traces

    def close_tracing_interface(self):
        """Called when someone closes the tracing interface to the device
        """

        pass

    def close_streaming_interface(self):
        """Called when someone closes the streaming interface to the device
        """

        self.stream_iface_open = False

    def push_script_chunk(self, chunk):
        """Called when someone pushes a new bit of a TRUB script to this device

        Args:
            chunk (str): a buffer with the next bit of script to append
        """

        self.script += bytearray(chunk)



"""IOTileReport subclass for readings packaged as individual readings
"""

import datetime
import struct
from report import IOTileReport, IOTileReading
from iotile.core.utilities.packed import unpack
from iotile.core.exceptions import ArgumentError, NotFoundError, EnvironmentError
import iotile.core.hw.auth.auth_provider as auth_provider
from iotile.core.hw.auth.auth_chain import ChainedAuthProvider

class SignedListReport(IOTileReport):
    """A report that consists of a signed list of readings.

    Args:
        rawreport (bytearray): The raw data of this report
    """

    ReportType = 1

    def __init__(self, rawreport, **kwargs):
        super(SignedListReport, self).__init__(rawreport, signed=True, encrypted=False, **kwargs)

    @classmethod
    def HeaderLength(cls):
        """Return the length of a header needed to calculate this report's length

        Returns:
            int: the length of the needed report header
        """

        return 20

    @classmethod
    def ReportLength(cls, header):
        """Given a header of HeaderLength bytes, calculate the size of this report
        """

        first_word, = unpack("<L", header[:4])

        length = (first_word >> 8)
        return length

    @classmethod
    def FromReadings(cls, uuid, readings, signature_method=auth_provider.HashOnlySHA256Signature, signer=None):
        """Generate an instance of the report format from a list of readings and a uuid

        Args:
            uuid (int): The uuid of the deviec that this report came from
            readings (list): A list of IOTileReading objects containing the data in the report
            signature_method (int): The method that should be used to sign the report (must be supported
                by an auth_provider)
            signer (AuthProvider): An optional preconfigured AuthProvider that should be used to sign this
                report.  If no AuthProvider is provided, the default ChainedAuthProvider is used.
        """

        lowest_id = IOTileReading.InvalidReadingID
        highest_id = IOTileReading.InvalidReadingID

        report_len = 20 + 16*len(readings) + 24
        len_low = report_len & 0xFF
        len_high = report_len >> 8

        unique_readings = [x.reading_id for x in readings if x.reading_id != IOTileReading.InvalidReadingID]
        if len(unique_readings) > 0:  
            lowest_id = min(unique_readings)
            highest_id = max(unique_readings)

        header = struct.pack("<BBHLLLBBH", cls.ReportType, len_low, len_high, uuid, 0, 0, signature_method, 0, 0xFFFF)
        header = bytearray(header)

        packed_readings = bytearray()

        for reading in readings:
            packed_reading = struct.pack("<HHLLL", reading.stream, 0, reading.reading_id, reading.raw_time, reading.value)
            packed_readings += bytearray(packed_reading)

        footer_stats = struct.pack("<LL", lowest_id, highest_id)
        signed_data = header + packed_readings + footer_stats

        if signer is None:
            signer = ChainedAuthProvider()

        try:
            signature = signer.sign(uuid, signature_method, signed_data)
        except NotFoundError:
            raise EnvironmentError("Could not sign report because no AuthProvider supported the requested signature method for the requested device", device_id=uuid, signature_method=signature_method)
        
        footer = struct.pack("16s", str(signature['signature'][:16]))
        footer = bytearray(footer)

        data = signed_data + footer
        return SignedListReport(data)

    def decode(self):
        """Decode this report into a list of readings
        """

        fmt, len_low, len_high, device_id, report_id, sent_timestamp, signature_flags, origin_streamer, streamer_selector = unpack("<BBHLLLBBH", self.raw_report[:20])

        assert fmt == 1
        length = (len_high << 8) | len_low

        self.origin = device_id
        self.report_id = report_id
        self.sent_timestamp = sent_timestamp
        self.origin_streamer = origin_streamer
        self.streamer_selector = streamer_selector
        self.signature_flags = signature_flags

        assert len(self.raw_report) == length

        remaining = self.raw_report[20:]
        assert len(remaining) >= 24
        readings = remaining[:-24]
        footer = remaining[-24:]

        lowest_id, highest_id, signature = unpack("<LL16s", footer)
        signature = bytearray(signature)

        signed_data = self.raw_report[:-16]
        signer = ChainedAuthProvider()

        try:
            verification = signer.verify(device_id, signature_flags, signed_data, signature)
            self.verified = verification['verified']
        except NotFoundError:
            pass

        self.signature = signature
        self.lowest_id = lowest_id
        self.highest_id = highest_id

        #Make sure this report has an integer number of readings
        assert (len(readings) % 16) == 0

        time_base = self.received_time - datetime.timedelta(seconds=sent_timestamp)
        parsed_readings = []

        for i in xrange(0, len(readings), 16):
            reading = readings[i:i+16]
            stream, _, reading_id, timestamp, value = unpack("<HHLLL", reading)

            parsed = IOTileReading(timestamp, stream, value, time_base=time_base, reading_id=reading_id)
            parsed_readings.append(parsed)

        #FIXME: attempt to validate the signature
        return parsed_readings

    def encode(self):
        """Turn this report into a serialized bytearray that could be decoded with a call to decode

        If we were generated from a binary report, just return that so that the signature remains intact,
        otherwise generate one and attempt to sign it 
        """

        if hasattr(self, 'raw_report'):
            return self.raw_report

        lowest_id = IOTileReading.InvalidReadingID
        highest_id = IOTileReading.InvalidReadingID

        report_len = 20 + 16*len(self.visible_readings) + 24
        len_low = report_len & 0xFF
        len_high = report_len >> 8

        if len(self.visible_readings) > 0:
            lowest_id = min([x.reading_id for x in self.visible_readings if x.reading_id != IOTileReading.InvalidReadingID])
            highest_id = max([x.reading_id for x in self.visible_readings if x.reading_id != IOTileReading.InvalidReadingID])

        header = struct.pack("<BBHLLLBBH", self.ReportType, len_low, len_high, self.origin, self.report_id, self.sent_timestamp, self.signature_flags, self.origin_streamer, self.streamer_selector)
        header = bytearray(header)

        readings = bytearray()

        for reading in self.visible_readings:
            packed_reading = struct.pack("<HHLLL", reading.stream, 0, reading.reading_id, reading.raw_time, reading.value)
            readings += bytearray(packed_reading)

        #FIXME: Actually calculate a footer here
        footer = struct.pack("<LL16s", lowest_id, highest_id, '\0'*16)
        footer = bytearray(footer)

        return header + readings + footer



from dict_verify import DictionaryVerifier
from list_verify import ListVerifier
from string_verify import StringVerifier
from int_verify import IntVerifier
from float_verify import FloatVerifier
from bool_verify import BooleanVerifier
from literal_verify import LiteralVerifier
from options_verify import OptionsVerifier
from verifier import Verifier



"""Manager for typed configuration variables

ConfigManager allows plugins to register typed config variables
and allows users to then set those variables.  Config variables
can optionally have default values.  Plugins to iotile-core can
register their own config variables under their own namespace
by registering an entry_point named 'iotile.config_variables'.

The object pointed to by the entry point should be a function that
returns a 2-tuple with a string and a list that maps a namespace prefix
to a list of 4-tuples with format:

[name, type, description, default_value]

If there is no default value, a 3-tuple can be used with just name, type
and description.

"""
import pkg_resources
import fnmatch
from collections import namedtuple
from iotile.core.dev.registry import ComponentRegistry
from iotile.core.exceptions import ArgumentError, EnvironmentError
from iotile.core.utilities.typedargs import context, param, return_type, stringable, type_system

MISSING = object()
ConfigVariable = namedtuple("ConfigVariable", ['name', 'type', 'description', 'default'])

@context("ConfigManager")
class ConfigManager(object):
    """A class for managing typed configuration variables

    ConfigManager can be used to querying which config variables are defined
    and to set or get the currently defined variables.
    """

    def __init__(self):
        self._known_variables = {}
        self._load_providers()
        self._load_functions()
        self._reg = ComponentRegistry()

    def _load_providers(self):
        """Load all config_variables providers using pkg_resources
        """

        for entry in pkg_resources.iter_entry_points('iotile.config_variables'):
            try:
                provider = entry.load()
                prefix, conf_vars = provider()
            except (ValueError, TypeError) as exc:
                raise EnvironmentError("Error loading config variables", package=entry.name, error=str(exc))

            for var in conf_vars:
                if len(var) != 3 and len(var) != 4:
                    raise EnvironmentError("Error loading config variable, invalid length", data=var, package=entry.name)

                name = prefix + ':' + var[0]
                if len(var) == 3:
                    var_obj = ConfigVariable(var[0], var[1], var[2], MISSING)
                else:
                    var_obj = ConfigVariable(name, var[1], var[2], var[3])

                if name in self._known_variables:
                    raise EnvironmentError("The same config variable was defined twice", name=name)

                self._known_variables[name] = var_obj

    def _load_functions(self):
        """Load all config functions that should be bound to this ConfigManager

        Config functions allow you to add functions that will appear under ConfigManager
        but call your specified function.  This is useful for adding complex configuration
        behavior that is callable from the iotile command line tool
        """

        for entry in pkg_resources.iter_entry_points('iotile.config_function'):
            try:
                conf_func = entry.load()
                name = conf_func.__name__

                self.add_function(name, conf_func)
            except (ValueError, TypeError) as exc:
                raise EnvironmentError("Error loading config function", name=name, error=str(exc))

    def _format_variable(self, name, var):
        """Format a helpful string describing a config variable

        Args:
            name (string): The prefixed name of the config variable
            var (ConfigVariable): the variable to format

        Returns:
            string: The formatted string in the form name (type): (default %s) description
        """

        if var.default is MISSING:
            return "%s (%s): %s [no default]" % (name, var.type, var.description)

        return "%s (%s): %s [default: %s]" % (name, var.type, var.description, var.default)

    @param("glob", "string", desc="Glob pattern for finding config variables")
    @return_type("list(string)")
    def list(self, glob):
        """List all matching config variables

        The glob parameter should be a wildcard expression like:
        build:* to find all config variables defined with a build prefix.

        Returns:
            string[]: A list of string descriptions containing descriptions and
                type information.
        """

        known_vars = [x for x in sorted(self._known_variables.iterkeys()) if fnmatch.fnmatchcase(x, glob)]
        return ['- ' + self._format_variable(x, self._known_variables[x]) for x in known_vars]

    @param("name", "string", desc="Config variable to find")
    @stringable
    def get(self, name):
        """Get the current value of a config variable
        """

        if name not in self._known_variables:
            raise ArgumentError("Unknown config variable", name=name)

        var = self._known_variables[name]

        try:
            val = self._reg.get_config(name)
        except ArgumentError:
            if var.default is not MISSING:
                val = var.default
            else:
                raise ArgumentError("Config variable not set and there is no default value", name=name)

        typed_val = type_system.convert_to_type(val, var.type)
        return typed_val

    @param("name", "string", desc="Config variable to find")
    @stringable
    def remove(self, name):
        """Remove any currently defined values for the named variable
        """

        self._reg.clear_config(name)

    @param("name", "string", desc="Config variable to set")
    @param("value", "string", desc="Value to set")
    def set(self, name, value):
        """Set the current avlue of a config variable
        """

        if name not in self._known_variables:
            raise ArgumentError("Unknown config variable", name=name)

        self._reg.set_config(name, value)

    @param("name", "string", desc="Config variable to find")
    @return_type("string")
    def describe(self, name):
        """Describe a config variable by name

        Returns:
            string: A short description of what the variable is used for
        """

        if name not in self._known_variables:
            raise ArgumentError("Unknown config variable", name=name)

        var = self._known_variables[name]
        return self._format_variable(name, var)

    def add_variable(self, name, var_type, desc, default=MISSING):
        """Add a temporary variable to the config variable manager

        This function is mainly useful for testing since it does not
        persistently store information about the variable.

        Args:
            name (string): The name of the variable
            var_type (string): The type of the variable.  This should be a type
                known to the type_system.
            desc (string): The description of what this variable is for
            default (string): An optional default value for the variable
        """

        self._known_variables[name] = ConfigVariable(name, var_type, desc, default)

    def add_function(self, name, callable):
        """Add a config function to the config variable manager

        Config functions are like config variables but are functions
        rather than variables.  Sometimes you want to expose configuration
        but you really need a function to actually do it.  For example,
        let's say you want to store a github OAUTH token on behalf of a
        user.  You would like a config function like `link_github` that
        walks the user through logging in to github and getting a token
        then you would like to save that token into the config manager like
        normal.

        add_function lets you bind a method to ConfigManager dynamically.
        The function should be a normal function with its first argument
        as self and it is turned into a bound method on this instance
        of config manager.

        Args:
            name (string): The attribute name for this function
            callable (callable): A function with first argument self
                that will be bound to this ConfigManager object as
                a method.
        """

        if hasattr(self, name):
            raise ArgumentError("Trying to add a Config function with a conflicting name", name=name)

        bound_callable = callable.__get__(self)
        setattr(self, name, bound_callable)



# This file is adapted from python code released by WellDone International
# under the terms of the LGPLv3.  WellDone International's contact information is
# info@welldone.org
# http://welldone.org
#
# Modifications to this file from the original created at WellDone International 
# are copyright Arch Systems Inc.

from __future__ import print_function
from decorator import decorator
import traceback
import sys

def deprecated(message=None):
	"""
	Decorator to mark a function as deprecated.  Anytime it is
	called will result in a backtrace being printed with the 
	deprecation message to stderr.
	"""

	if message is None:
		message = "This function has been deprecated."
	if not isinstance(message, basestring):
		message = str(message)

	def _deprecated(f, *args, **kwargs):
		print("Function '%s' has been deprecated" % f.__name__, file=sys.stderr)
		print("Reason: %s" % message, file=sys.stderr)
		print("Callstack follows", file=sys.stderr)
		traceback.print_stack(file=sys.stderr)

		return f(*args, **kwargs)

	def wrap_func(f):
		return decorator(_deprecated, f)

	return wrap_func


# This file is copyright Arch Systems, Inc.  
# Except as otherwise provided in the relevant LICENSE file, all rights are reserved.

# external_proxy.py
# Routines for importing proxy modules from registered components on your computer

import pkg_resources
from iotile.core.dev.registry import ComponentRegistry
from proxy import TileBusProxyObject
from plugin import TileBusProxyPlugin
from iotile.core.exceptions import *
import os
import imp
import inspect
import sys

def find_proxy(component, proxy_name=None, obj_type=TileBusProxyObject):
    """
    Search through the registered component data base for a proxy or proxy plugin module provided by
    a specific component optionally having a specific name.  

    Proxy modules must inherit from TileBusProxyObject. Proxy plugin modules must inherit from TileBusProxyPlugin
    """

    # Find all of the registered IOTile components and see if we need to add any proxies for them
    reg = ComponentRegistry()
    comp = reg.find_component(component)

    proxies = comp.proxy_modules()
    if len(proxies) != 1:
        raise DataError("Attempting to find proxy module from component that does not provide exactly 1 proxy", proxies=proxies)

    proxy_mod = proxies[0]

    proxy_objs = import_proxy(proxy_mod, obj_type)

    if len(proxy_objs) == 0:
        raise DataError("Specified component did not define any proxies but said that it did provide them", component=component)
    elif len(proxy_objs) > 1 and proxy_name is None:
        raise DataError("Sepcific component defined multiple proxy objects and a name was not specified", component=component, proxies=proxy_objs)
    elif len(proxy_objs) == 1 and proxy_name is None:
        return proxy_objs[0]

    for obj in proxy_objs:
        if obj.__name__ == proxy_name:
            return obj

    raise DataError("Named proxy could not be found", component=component, proxies=proxy_objs, desired_name=proxy_name)

def find_proxy_plugin(component, plugin_name):
    """ Attempt to find a proxy plugin provided by a specifc component
    
    Args:
        component (string): The name of the component that provides the plugin
        plugin_name (string): The name of the plugin to load

    Returns:
        TileBuxProxPlugin: The plugin, if found, otherwise raises DataError
    """

    reg = ComponentRegistry()

    #Try to find plugin in an installed component, otherwise look through installed
    #packages
    try:
        comp = reg.find_component(component)

        plugins = comp.proxy_plugins()

        for plugin in plugins:
            objs = import_proxy(plugin, TileBusProxyPlugin)
            for obj in objs:
                if obj.__name__ == plugin_name:
                    return obj
    except ArgumentError:
        pass

    for entry in pkg_resources.iter_entry_points('iotile.proxy_plugin'):
        module = entry.load()
        objs = [obj for obj in module.__dict__.itervalues() if inspect.isclass(obj) and issubclass(obj, TileBusProxyPlugin) and obj != TileBusProxyPlugin]
        for obj in objs:
            if obj.__name__ == plugin_name:
                return obj

    raise DataError("Could not find proxy plugin module in registered components or installed distributions", component=component, name=plugin_name)

def import_proxy(path, obj_type):
    """
    Add all proxy objects defined in the python module located at path.

    The module is loaded and all classes that inherit from the given object type
    are loaded and can be used to interact later with modules of that type.

    Returns the total number of proxies added.
    """

    d,p = os.path.split(path)

    p,ext = os.path.splitext(p)
    if ext != '.py' and ext != '.pyc' and ext != "":
        raise ArgumentError("Passed proxy module is not a python package or module (.py or .pyc)", path=path)

    try:
        fileobj,pathname,description = imp.find_module(p, [d])

        #Don't import twice if we've already imported this module
        if p in sys.modules:
            mod = sys.modules[p]
        else:
            mod = imp.load_module(p, fileobj, pathname, description)
    except ImportError as e:
        raise ArgumentError("could not import module in order to load external proxy modules", module_path=path, parent_directory=d, module_name=p, error=str(e))

    num_added = 0
    return [obj for obj in filter(lambda x: inspect.isclass(x) and issubclass(x, obj_type) and x != obj_type, mod.__dict__.itervalues())]



"""IOTileReport subclass for readings packaged as individual readings
"""

import datetime
import struct
from report import IOTileReport, IOTileReading
from iotile.core.utilities.packed import unpack
from iotile.core.exceptions import ArgumentError


class IndividualReadingReport(IOTileReport):
    """Report format where every reading is packaged in a 20 byte frame

    Args:
        rawreport (bytearray): The raw data of this report
    """

    ReportType = 0

    def __init__(self, rawreport, **kwargs):
        super(IndividualReadingReport, self).__init__(rawreport, signed=False, encrypted=False, **kwargs)

    @classmethod
    def HeaderLength(cls):
        """Return the length of a header needed to calculate this report's length

        Returns:
            int: the length of the needed report header
        """

        return 1

    @classmethod
    def ReportLength(cls, header):
        """Given a header of HeaderLength bytes, calculate the size of this report
        """

        return 20

    @classmethod
    def FromReadings(cls, uuid, readings):
        """Generate an instance of the report format from a list of readings and a uuid
        """

        if len(readings) != 1:
            raise ArgumentError("IndividualReading reports must be created with exactly one reading", num_readings=len(readings))

        reading = readings[0]
        data = struct.pack("<BBHLLLL", 0, 0, reading.stream, uuid, 0, reading.raw_time, reading.value)
        return IndividualReadingReport(data)

    def decode(self):
        """Decode this report into a single reading
        """

        fmt, _, stream, uuid, sent_timestamp, reading_timestamp, reading_value = unpack("<BBHLLLL", self.raw_report)
        assert fmt == 0

        #Estimate the UTC time when this device was turned on
        time_base = self.received_time - datetime.timedelta(seconds=sent_timestamp)

        reading = IOTileReading(reading_timestamp, stream, reading_value, time_base=time_base)
        self.origin = uuid
        self.sent_timestamp = sent_timestamp

        return [reading]

    def encode(self):
        """Turn this report into a serialized bytearray that could be decoded with a call to decode
        """

        reading = self.visible_readings[0]
        data = struct.pack("<BBHLLLL", 0, 0, reading.stream, self.origin, self.sent_timestamp, reading.raw_time, reading.value)

        return bytearray(data)


from verifier import Verifier
from iotile.core.exceptions import ValidationError


class BooleanVerifier(Verifier):
    """Verify that an object is a boolean

    Args:
        desc (string): A description of the verifier
        require_value (bool): Require a specific True or False value
    """

    def __init__(self, require_value=None, desc=None):
        super(BooleanVerifier, self).__init__(desc)
        self._require_value = require_value

    def verify(self, obj):
        """Verify that the object conforms to this verifier's schema

        Args:
            obj (object): A python object to verify

        Raises:
            ValidationError: If there is a problem verifying the dictionary, a
                ValidationError is thrown with at least the reason key set indicating
                the reason for the lack of validation.
        """

        if not isinstance(obj, bool):
            raise ValidationError("Object is not a bool", reason='object is not a bool', object=obj)

        if self._require_value is not None and obj != self._require_value:
            raise ValidationError("Boolean is not equal to specified literal", reason='boolean value %s should be %s' % (str(obj), str(self._require_value)))

    def format(self, indent_level, indent_size=4):
        """Format this verifier

        Returns:
            string: A formatted string
        """

        name = self.format_name('Boolean', indent_size)

        if self._require_value is not None:
            if self.long_desc is not None:
                name += '\n'

            name += self.wrap_lines('must be %s\n' % str(self._require_value).lower(), 1, indent_size)

        return self.wrap_lines(name, indent_level, indent_size)



# This file is adapted from python code released by WellDone International
# under the terms of the LGPLv3.  WellDone International's contact information is
# info@welldone.org
# http://welldone.org
#
# Modifications to this file from the original created at WellDone International 
# are copyright Arch Systems Inc.

# IOTile Exceptions

class IOTileException(Exception):
    """
    All MoMo API routines, upon encountering unrecoverable errors,
    will throw an exception that is a subclass of this exception.
    All other exceptions are caught and rethrown as an APIError
    for consistency.  Calling methods can assume that all annotated
    API functions will throw only IOTileException subclasses as defined
    below.
    """

    def __init__(self, msg, **kwargs):
        self.params = kwargs
        self.msg = msg
    
    def format(self):
        msg = "%s: %s" % (self.__class__.__name__, self.msg)

        if len(self.params) != 0:
            paramstring = "\n".join([str(key) + ": " + str(val) for key,val in self.params.iteritems()])
            msg += "\nAdditional Information:\n" + paramstring
        
        return msg

    def format_msg(self):
        msg = self.msg
        if len(self.params) != 0:
            paramstring = "\n".join([str(key) + ": " + str(val) for key,val in self.params.iteritems()])
            msg += "\nAdditional Information:\n" + paramstring

        return msg

    def to_dict(self):
        """Convert this excpetion to a dictionary

        Has a 'reason' key and a dictionary of params
        """

        out = {}
        out['reason'] = self.msg
        out['type'] = self.__class__.__name__
        out['params'] = self.params

        return out

    def __str__(self):
        msg = self.format()
        return msg

class ValidationError(IOTileException):
    """
    API routines can impose validation criteria on` their arguments in 
    addition to requiring simply a certain type of argument.  A clasic
    example is the "path" type which can have validators like "readable"
    or "exists".  When validation fails, this error is thrown.
    """

    pass

class ConversionError(IOTileException):
    """
    All API functions take typed parameters.  Each type defines conversion 
    operators for python types that are logically related to it.  When no
    valid conversion exists for the data type passed, this error is thrown.
    """

    pass

class NotFoundError(IOTileException):
    """
    Thrown when an attempt to execute an API method is made and the API
    method does not exist.
    """

    pass

class TimeoutError(IOTileException):
    """
    The method timed out, usually indicating that a communication failure
    occurred, either with another process or with a hardware module.
    """

    pass

class ArgumentError(IOTileException):
    """
    The method could not be called with the arguments passed.  This
    differs from InternalError in that when ArgumentError is returned,
    it is known that these arguments can never work for this function.

    An example would be passing three arguments to a function requiring
    4 arguments.
    """

    pass

class DataError(IOTileException):
    """
    The method relied on data pass in by the user and the data was invalid.

    This could be because a file was the wrong type or because a data provider
    returned an unexpected result.  The parameters passed with this exception
    provide more detail on what occurred and where.
    """

    pass

class InternalError(IOTileException):
    """
    The method could not be completed with the user input passed for 
    an unexpected reason.  This does not signify a bug in the API 
    method code.  More details should be passed in the arguments.
    """

    pass

class APIError(IOTileException):
    """
    An internal API error occured during the execution of the method.
    This should only be returned if the error was unforeseen and not
    caused in any way by user input.  If the problem is that a user 
    input is invalid for the API call, ValidationError should be
    thrown instead.  

    All instances of APIError being thrown are bugs that should be
    reported and fixed. 
    """

    pass

class BuildError(IOTileException):
    """
    There is an error with some part of the build system.  This does not
    mean that there is a compilation error but rather that a required part
    of the build process did not complete successfully.  This exception means
    that something is misconfigured.
    """

    pass

class TypeSystemError(IOTileException):
    """
    There was an error with the MoMo type system.  This can be due to improperly
    specifying an unknown type or because the required type was not properly loaded
    from an external module before a function that used that type was needed.
    """

    pass

class EnvironmentError(IOTileException):
    """
    The environment is not properly configured for the MoMo API command that was called.
    This can be because a required program was not installed or accessible or because
    a required environment variable was not defined.
    """

    pass

class HardwareError(IOTileException):
    """
    There was an issue communicating with or controlling a MoMo hardware module.  This
    exception anchors a range of exceptions that refer to specific kinds of hardware issues.

    By catching this exception, you will catch any sort of hardware failure.  If you are
    interested in specific kinds of hardware errors, you can look for or catch subclasses
    of this exception.
    """

    pass



from verifier import Verifier
from iotile.core.exceptions import ValidationError


class ListVerifier(Verifier):
    """Verify the values in a List

    Args:
        desc (string): A description of what this list should contain
        verifier (Verifier): A verifier for what the list should contain.
            each entry in the list must match the same verifier.
        min_length (int): An optional minimum length for the list
        max_length (int): An optional maximum length for the list
    """

    def __init__(self, verifier, min_length=None, max_length=None, desc=None):
        super(ListVerifier, self).__init__(desc)

        self._min_length = min_length
        self._max_length = max_length
        self._verifier = verifier

    def verify(self, obj):
        """Verify that the object conforms to this verifier's schema

        Args:
            obj (object): A python object to verify

        Raises:
            ValidationError: If there is a problem verifying the dictionary, a
                ValidationError is thrown with at least the reason key set indicating
                the reason for the lack of validation.
        """

        if self._min_length is not None and len(obj) < self._min_length:
            raise ValidationError("List was too short", reason="list length %d was less than the minimum %d" % (len(obj), self._min_length), min_length=self._min_length, actual_length=len(obj))

        if self._max_length is not None and len(obj) > self._max_length:
            raise ValidationError("List was too long", reason="list length %d was greater than the maximum %d" % (len(obj), self._max_length), min_length=self._max_length, actual_length=len(obj))

        for i, val in enumerate(obj):
            self._verifier.verify(val)
